/*
 * tmos - thread.c
 * 	version 0.0.0
 *
 * 	TMOS multitasking demo
 * 
 * Copyright (C) 2005..2010  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <asm/system.h>
#include <tmos/sched.h>
#include <tmos/port.h>
#include <tmos/mutex.h>
#include <tmos/jiffies.h>

/* console port */
static port_t port;

/* mutex to protect console from concurent access */
static DECLARE_MUTEX(put_lock);

/**
 * __putc - print character to console
 * @c: character to print
 *
 * This function must be called from kernel mode with console lock held.
 */
static void __putc(char c) __reentrant
{
	port_write(port, &c, 1);
}

/**
 * putc - print character to console
 * @c: character to print
 *
 * This function must be called from user mode.
 */
static void putc(char c) __reentrant
{
	mutex_lock(put_lock);
	__putc(c);
	mutex_unlock(put_lock);
}

/**
 * puts - print string to console
 * @s: string to print
 *
 * This function must be called from user mode.
 */
static void puts(const char *s) __reentrant
{
	while (*s)
		putc(*s++);
}

/**
 * task_a_f - task A code
 */
static void task_a_f(void) __reentrant
{
	tsc_t ht;

	/* child code */
	for (;;) {
		ht = tsc_read();
		schedule();
		ht = tsc_read() - ht;

		printk("sched: %d\n", ht);
	}
}

void app_main(void) __reentrant
{
	printk("openning console\n");
	port = port_open(CONFIG_CONSOLE_DEVICE, O_RDWR);
	if (port == -1)
		BUG();
	
	puts("testing puts.\n");

	printk("creating task A\n");
	if (!tfork()) {
		task_a_f();
		return;
	}

	for (;;) {
		schedule();
	}
}
