/*
 * tmos - context.c
 *
 * Copyright (C) 2009  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <tmos/kernel.h>
#include <tmos/serial.h>
#include <tmos/jiffies.h>
#include <asm/task.h>
#include <asm/8052.h>

struct context __kernel *current = KERNEL_NULL;

#if CONFIG_PREEMPT
char preempt_count = 0;
#endif

/**
 * context_init - initialize context
 * @context: context to initialize
 */
char context_init(struct context __kernel *context) __preempt
{
#ifdef CONFIG_TASK_TIMES
	context->cpu_time = 0;
#endif
	return 0;
}

#if CONFIG_TDATA
#define __XPUSH(x) \
	do { \
		*xstack = (x); \
	} while (0)

#define XPUSH(x) \
	do { \
		xstack++; sp++; \
		__XPUSH(x); \
	} while (0)
#else
#define __XPUSH(x) \
	do { \
		*((unsigned char __data *) sp) = (x); \
	} while (0)

#define XPUSH(x) \
	do { \
		sp++; \
		__XPUSH(x); \
	} while (0)
#endif
	
#define USER_R(x) (*((unsigned char __data *) (x)))
#define USER_R0 USER_R(0)
#define USER_R1 USER_R(1)
#define USER_R2 USER_R(2)
#define USER_R3 USER_R(3)
#define USER_R4 USER_R(4)
#define USER_R5 USER_R(5)
#define USER_R6 USER_R(6)
#define USER_R7 USER_R(7)

#ifndef CONFIG_PREEMPT
/**
 * thread_create_child - new thread startup
 *
 * The child thread starts here, we must return 1.
 *
 * This function is not needed in preemptive kernel because in
 * preemptive kernel we can create proper register dump with
 * 1 in DPL register.
 */
static char thread_create_child(void)
{
	return 1;
}
#endif

unsigned int foo(int a, unsigned char b, unsigned int c)
{
	return c;
}

extern unsigned char bp;

/**
 * thread_create - create new thread with stack at sp
 * @context: thread to create
 * @sp: thread stack pointer
 *
 * This function returns 0 on parent thread and 1 on child thread (identified
 * with context).
 *
 * This function must be called with preemption disabled.
 *
 * This function is fragile. Don't change it.
 */
char __thread_create(struct context __kernel *context, unsigned char sp,
		unsigned int entry)
	__need_lock(PREEMPT) __reentrant
{
#if CONFIG_TDATA
	register unsigned char __xdata *xstack;

	xstack = (unsigned char __xdata *)
		((task_tdata(context) << 8) + sp - TDATA_BASE);
#endif
	__XPUSH((u8) entry);
	XPUSH(entry >> 8);

#ifndef CONFIG_PREEMPT
	XPUSH(((unsigned int) thread_create_child) & 0xff);
	XPUSH(((unsigned int) thread_create_child) >> 8);
#endif

	XPUSH(0);	/* Acc */
	XPUSH(B);	/* B */
	XPUSH(bp);	/* BP */
	XPUSH(PSW);	/* PSW */

#if CONFIG_PREEMPT
#if WANT_PUSH_DP
	XPUSH(1);	/* DPL */
	XPUSH(DPH);	/* DPH */
#endif
	XPUSH(USER_R0);	/* R0 */
	XPUSH(USER_R1);	/* R1 */
	XPUSH(USER_R2);	/* R2 */
	XPUSH(USER_R3);	/* R3 */
	XPUSH(USER_R4);	/* R4 */
	XPUSH(USER_R5);	/* R5 */
	XPUSH(USER_R6);	/* R6 */
	XPUSH(USER_R7);	/* R7 */
#endif

#ifdef CONFIG_DEBUG
	XPUSH(0xfa);	/* POISSON */
	sp--;
#endif

	context->sp = sp;
	return 0;
}

#if 0
char thread_create(struct context __kernel *context, unsigned char sp)
	__need_lock(PREEMPT) __reentrant __naked
{
	for (;;);
}
#endif

#if 1
char thread_create(struct context __kernel *context, unsigned char sp)
	__need_lock(PREEMPT) __reentrant __naked
{
#ifdef __INLINE_ASM
	/*
	 * Before		After:
	 * sp
	 * ret_hi		ret_hi
	 * ret_lo		ret_lo
	 * 			sp
	 * 			[ret]
	 * 			[ret]
	 * 		ar0
	 * 		ar1
	 */
	setup_registers(0);
	__asm
		mov	acc, sp
		inc	sp
		push	ar0
		push	ar1
		mov	r1, a
		add	a, #0xfe
		mov	r0, a

		inc	r1

		/* copy sp */
		mov	a, @r0
		mov	@r1, a

		pop	ar1
		pop	ar0
		lcall	___thread_create
		dec	sp
		ret
	__endasm;
#endif
}
#endif

extern void softirq(void);
void show_context_state(struct context __kernel *context)
{
	__dump_stack(context->sp);
}
