/*
 * tmos - processor.h
 *
 * Copyright (C) 2009  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __ASM_PROCESSOR_H
#define __ASM_PROCESSOR_H

#include <tmos/config.h>

/**
 * cpu_relax() should be inserted in busy-wait loops. cpu_relax()
 * is at least compiler barrier to avoid optimizing out busy-wait loops.
 * It can also perform some power savings or resource savings.
 */
#define cpu_relax() barrier()

#define nop() __asm__(NOP)

#ifdef CONFIG_USE_IDLE
#define halt() \
	do { \
		barrier(); \
		PCON |= IDL; \
		barrier(); \
	} while (0)
#else
#define halt()		cpu_relax()
#endif

#ifdef CONFIG_USE_IDLE
/*
 * safe_halt - raceless halt(), must be called with interrupts disabled
 *
 * nop() is required because some 8051s might execute one instruction
 * after wakeup caused by RESET.
 */
#define safe_halt() \
	do { \
		barrier(); \
		local_irq_enable(); \
		PCON |= IDL; \
		nop(); \
		barrier(); \
	} while (0)
#else
#define safe_halt()		local_irq_enable()
#endif

#endif
